package ca.yorku.cse.mack.tiltmaze;

import android.app.Activity;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.Bundle;
import android.view.View;
import android.widget.ArrayAdapter;
import android.widget.CheckBox;
import android.widget.Spinner;
import android.widget.Toast;

public class SetupActivity extends Activity 
{
	// keys for setup preferences
	final static String PARTICIPANT_CODE_KEY = "participant_code";
	final static String SESSION_CODE_KEY = "session_code";
	final static String GROUP_CODE_KEY = "group_code";
	final static String TRIALS_CODE_KEY = "trials_code";
	final static String ORDER_OF_CONTROL_KEY = "orderOfControl";
	final static String GAIN_KEY = "gain";
	final static String GAIN_LABEL_KEY = "gain_label";
	final static String SHOW_NOMINAL_PATH_KEY = "show_nominal_path";
	final static String SHOW_MINIMUM_PATH_KEY = "show_minimum_path";
	final static String SHOW_WALL_HITS_KEY = "show_wall_hits";
	final static String SOUND_ALARM_ON_WALL_HIT_KEY = "sound_alarm_on_wall_hit";

	// setup parameters (1st entry is overwritten from values saved in shared preferences, if any)
	String[] participantCode = { "P99", "P01", "P02", "P03", "P04", "P05", "P06", "P07", 
			"P08", "P09", "P10", "P11", "P12", "P13", "P14", "P15", "P16" };
	String[] sessionCode = { "S99", "S01", "S02", "S03", "S04", "S05" };	// T for trials and S for sessions
	String[] blockCode = { "(auto)" }; // effectively disable spinner for block code
	String[] groupCode = { "G99", "G01", "G02", "G03", "G04" };	// Determines which Order of Control participant completed first
	String[] trialsCode = { "1", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "15", "20", "25" };
	String[] order_of_control = { "Velocity", "Velocity", "Position" };
	String[] gain = { "Medium", "Very low", "Low", "Medium", "High", "Very high" };

	boolean showNominalPath = false;
	boolean showMinimumPath = false;
	boolean showWallHits = false;
	boolean soundAlarmOnWallHit = false;
	
	// somewhat arbitrary mappings for gain by order of control
	final static int[] GAIN_ARG_POSITION_CONTROL = { 20, 5, 10, 20, 40, 80 };
	final static int[] GAIN_ARG_VELOCITY_CONTROL = { 100, 25, 50, 100, 200, 400 };

	// save settings
	SharedPreferences sp;
	SharedPreferences.Editor spe;

	// spinners
	private Spinner spinParticipant, spinSession, spinBlock, spinGroup, spinTrials, spinOrderOfControl, spinGain;
	private CheckBox checkShowNominalPath, checkShowMinimumPath, checkShowWallHits, checkSoundAlarmOnWallHit;

	// called when the activity is first created
	@Override
	public void onCreate(Bundle savedInstanceState) 
	{
		super.onCreate(savedInstanceState);
		setContentView(R.layout.setup);
		
		/*
		 * Initialize reference to shared preferences. NOTE: The values saved are the default
		 * values.
		 */
		sp = this.getPreferences(MODE_PRIVATE);

		/*
		 * Overwrite default (1st) entry from setup parameters, if a different value has been
		 * previously saved.
		 */
		participantCode[0] = sp.getString(PARTICIPANT_CODE_KEY, participantCode[0]);
		sessionCode[0] = sp.getString(SESSION_CODE_KEY, sessionCode[0]);
		// block code initialised in main activity (based on existing filenames)
		groupCode[0] = sp.getString(GROUP_CODE_KEY, groupCode[0]);
		trialsCode[0] = sp.getString(TRIALS_CODE_KEY, trialsCode[0]);
		order_of_control[0] = sp.getString(ORDER_OF_CONTROL_KEY, order_of_control[0]);
		gain[0] = sp.getString(GAIN_KEY, gain[0]);
		showNominalPath = sp.getBoolean(SHOW_NOMINAL_PATH_KEY, showNominalPath);
		showMinimumPath = sp.getBoolean(SHOW_MINIMUM_PATH_KEY, showMinimumPath);
		showWallHits = sp.getBoolean(SHOW_WALL_HITS_KEY, showWallHits);
		soundAlarmOnWallHit = sp.getBoolean(SOUND_ALARM_ON_WALL_HIT_KEY, soundAlarmOnWallHit);

		// get references to widget elements
		spinParticipant = (Spinner)findViewById(R.id.paramPart);
		spinSession = (Spinner)findViewById(R.id.paramSess);
		spinBlock = (Spinner)findViewById(R.id.paramBlock);
		spinGroup = (Spinner)findViewById(R.id.paramGroup);
		spinTrials = (Spinner)findViewById(R.id.paramTrials);
		spinOrderOfControl = (Spinner) findViewById(R.id.paramOrderOfControl);
		spinGain = (Spinner) findViewById(R.id.paramGain);
		checkShowNominalPath = (CheckBox)findViewById(R.id.paramShowNominalPath);
		checkShowMinimumPath = (CheckBox)findViewById(R.id.paramShowMinimumPath);
		checkShowWallHits = (CheckBox)findViewById(R.id.paramShowWallHits);
		checkSoundAlarmOnWallHit = (CheckBox)findViewById(R.id.paramSoundAlarmOnWallHit);
   
		// initialise spinner adapters
		ArrayAdapter<CharSequence> adapterPC = new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle,
				participantCode);
		spinParticipant.setAdapter(adapterPC);

		ArrayAdapter<CharSequence> adapterSS = new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, sessionCode);
		spinSession.setAdapter(adapterSS);

		ArrayAdapter<CharSequence> adapterB = new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, blockCode);
		spinBlock.setAdapter(adapterB);

		ArrayAdapter<CharSequence> adapterG = new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, groupCode);
		spinGroup.setAdapter(adapterG);

		ArrayAdapter<CharSequence> adapterT = new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, trialsCode);
		spinTrials.setAdapter(adapterT);

		ArrayAdapter<CharSequence> adapter2 = new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, order_of_control);
		spinOrderOfControl.setAdapter(adapter2);
		
		ArrayAdapter<CharSequence> adapterGN= new ArrayAdapter<CharSequence>(this, R.layout.spinnerstyle, gain);
		spinGain.setAdapter(adapterGN);
		
		// initialize checkbox
		checkShowNominalPath.setChecked(showNominalPath);
		checkShowMinimumPath.setChecked(showMinimumPath);
		checkShowWallHits.setChecked(showWallHits);
		checkSoundAlarmOnWallHit.setChecked(soundAlarmOnWallHit);
	}

	// called when the "OK" button is pressed
	public void clickOK(View view) 
	{
		// get user's choices... 
		String part = participantCode[spinParticipant.getSelectedItemPosition()];
		String sess = sessionCode[spinSession.getSelectedItemPosition()];
		// block handled differently (auto generated from existing filenames)
		String group = groupCode[spinGroup.getSelectedItemPosition()];
		int numberOfTrials = Integer.parseInt(trialsCode[spinTrials.getSelectedItemPosition()]);		
		String orderOfControl = (String) spinOrderOfControl.getSelectedItem();
		boolean showNominalPath = checkShowNominalPath.isChecked();
		boolean showMinimumPath = checkShowMinimumPath.isChecked();
		boolean showWallHits = checkShowWallHits.isChecked();
		boolean soundAlarmOnWallHit = checkSoundAlarmOnWallHit.isChecked();
		
		// save the user's choices
		spe = sp.edit();
		spe.putString(PARTICIPANT_CODE_KEY, participantCode[spinParticipant.getSelectedItemPosition()]);
		spe.putString(SESSION_CODE_KEY, sessionCode[spinSession.getSelectedItemPosition()]);
		spe.putString(GROUP_CODE_KEY, groupCode[spinGroup.getSelectedItemPosition()]);
		spe.putString(TRIALS_CODE_KEY, trialsCode[spinTrials.getSelectedItemPosition()]);
		spe.putString(ORDER_OF_CONTROL_KEY, order_of_control[spinOrderOfControl.getSelectedItemPosition()]);
		spe.putString(GAIN_KEY, gain[spinGain.getSelectedItemPosition()]);
		spe.putBoolean(SHOW_NOMINAL_PATH_KEY, checkShowNominalPath.isChecked());
		spe.putBoolean(SHOW_MINIMUM_PATH_KEY, checkShowMinimumPath.isChecked());
		spe.putBoolean(SHOW_WALL_HITS_KEY, checkShowWallHits.isChecked());
		spe.putBoolean(SOUND_ALARM_ON_WALL_HIT_KEY, checkSoundAlarmOnWallHit.isChecked());
		spe.commit();

		// actual gain value depends on order of control
		int gain;
		if (orderOfControl.equals("Velocity"))
			gain = GAIN_ARG_VELOCITY_CONTROL[spinGain.getSelectedItemPosition()];
		else
			gain = GAIN_ARG_POSITION_CONTROL[spinGain.getSelectedItemPosition()];
		
		String gainLabel = (String)spinGain.getSelectedItem();

		// Laps
		//int numLaps = 1;

		// bundle up parameters to pass on to activity
		Bundle b = new Bundle();
		b.putString(PARTICIPANT_CODE_KEY, part);
		b.putString(SESSION_CODE_KEY, sess);
		b.putString(GROUP_CODE_KEY, group);
		b.putInt(TRIALS_CODE_KEY, numberOfTrials);
		b.putString(ORDER_OF_CONTROL_KEY, orderOfControl);
		b.putInt(GAIN_KEY, gain);
		b.putString(GAIN_LABEL_KEY, gainLabel);
		b.putBoolean(SHOW_NOMINAL_PATH_KEY, showNominalPath);
		b.putBoolean(SHOW_MINIMUM_PATH_KEY, showMinimumPath);
		b.putBoolean(SHOW_WALL_HITS_KEY, showWallHits);
		b.putBoolean(SOUND_ALARM_ON_WALL_HIT_KEY, soundAlarmOnWallHit);

		// start experiment activity
		Intent i = new Intent(getApplicationContext(), TiltMazeActivity.class);
		i.putExtras(b);
		startActivity(i);
		finish();
	}

	/** Called when the "Exit" button is pressed. */
	public void clickExit(View view) 
	{
		super.onDestroy(); // cleanup
		this.finish(); // terminate
	}
}
